//
//  ViewController.swift
//  DSWaveformImageExample
//
//  Created by Dennis Schmidt on 06/02/2017.
//  Copyright © 2017 Dennis Schmidt. All rights reserved.
//

import UIKit
import DSWaveformImage
import DSWaveformImageViews

class ViewController: UIViewController {
    @IBOutlet weak var topWaveformView: UIImageView!
    @IBOutlet weak var middleWaveformView: WaveformImageView!
    @IBOutlet weak var bottomWaveformView: UIImageView!

    private let waveformImageDrawer = WaveformImageDrawer()
    private let audioURL = Bundle.main.url(forResource: "example_sound", withExtension: "m4a")!

    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)

        updateWaveformImages()

        // get access to the raw, normalized amplitude samples
        let waveformAnalyzer = WaveformAnalyzer(audioAssetURL: audioURL)
        waveformAnalyzer?.samples(count: 10) { samples in
            print("sampled down to 10, results are \(samples ?? [])")
        }
    }

    override func viewDidLayoutSubviews() {
        super.viewDidLayoutSubviews()
        updateWaveformImages()
    }

    private func updateWaveformImages() {
        // always uses background thread rendering
        waveformImageDrawer.waveformImage(
            fromAudioAt: audioURL, with: .init(
                size: topWaveformView.bounds.size,
                style: .gradient(
                    [
                        UIColor(red: 255/255.0, green: 159/255.0, blue: 28/255.0, alpha: 1),
                        UIColor(red: 255/255.0, green: 191/255.0, blue: 105/255.0, alpha: 1),
                        UIColor.red
                    ]
                ),
                dampening: .init(percentage: 0.2, sides: .right, easing: { x in pow(x, 4) }),
                position: .top,
                verticalScalingFactor: 2)
        ) { image in
            // need to jump back to main queue
            DispatchQueue.main.async {
                self.topWaveformView.image = image
            }
        }

        middleWaveformView.configuration = Waveform.Configuration(
            backgroundColor: .lightGray.withAlphaComponent(0.1),
            style: .striped(.init(color: UIColor(red: 51/255.0, green: 92/255.0, blue: 103/255.0, alpha: 1), width: 5, spacing: 5)),
            verticalScalingFactor: 0.5
        )
        middleWaveformView.waveformAudioURL = audioURL

        // basic illustration of async / await API usage
        Task {
            let image = try! await waveformImageDrawer.waveformImage(fromAudioAt: audioURL, with: bottomWaveformConfiguration)

            // as an added bonus, use CALayer's compositingFilter for more elaborate image display
            self.bottomWaveformView.layer.compositingFilter = "multiplyBlendMode"
            self.bottomWaveformView.image = image
        }

        // Photo by Alexander Popov on Unsplash
        // https://unsplash.com/@5tep5?utm_source=unsplash&utm_medium=referral&utm_content=creditCopyText
        // https://unsplash.com/s/photos/techno?utm_source=unsplash&utm_medium=referral&utm_content=creditCopyText
    }

    private var bottomWaveformConfiguration: Waveform.Configuration {
        Waveform.Configuration(
            size: bottomWaveformView.bounds.size,
            style: .filled(UIColor(red: 129/255.0, green: 178/255.0, blue: 154/255.0, alpha: 1)),
            position: .middle
        )
    }
}
